# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function

import glob
import logging
import os.path

from reno import loader
from reno import scanner

LOG = logging.getLogger(__name__)


def lint_cmd(args, conf):
    "Check some common mistakes"
    LOG.debug('starting lint')
    notesdir = os.path.join(conf.reporoot, conf.notespath)
    notes = glob.glob(os.path.join(notesdir, '*.yaml'))

    error = 0
    load = loader.Loader(conf, ignore_cache=True)
    allowed_section_names = [conf.prelude_section_name] + \
                            [s[0] for s in conf.sections]

    uids = {}
    for f in notes:
        LOG.debug('examining %s', f)
        uid = scanner._get_unique_id(f)
        uids.setdefault(uid, []).append(f)

        content = load.parse_note_file(f, None)
        for section_name in content.keys():
            if section_name not in allowed_section_names:
                LOG.warning('unrecognized section name %s in %s',
                            section_name, f)
                error = 1

    for uid, names in sorted(uids.items()):
        if len(names) > 1:
            LOG.warning('UID collision: %s', names)
            error = 1

    return error



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function

from reno import formatter
from reno import loader


def report_cmd(args, conf):
    "Generates a release notes report"
    ldr = loader.Loader(conf)
    if args.version:
        versions = args.version
    else:
        versions = ldr.versions
    text = formatter.format_report(
        ldr,
        conf,
        versions,
        title=args.title,
        show_source=args.show_source,
    )
    if args.output:
        with open(args.output, 'w') as f:
            f.write(text)
    else:
        print(text)
    return



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import binascii
import logging
import os
import os.path
import random
import subprocess

LOG = logging.getLogger(__name__)


def get_random_string(nbytes=8):
    """Return a fixed-length random string

    :rtype: six.text_type
    """
    try:
        # NOTE(dhellmann): Not all systems support urandom().
        # hexlify returns six.binary_type, decode to convert to six.text_type.
        val = binascii.hexlify(os.urandom(nbytes)).decode('utf-8')
    except Exception as e:
        print('ERROR, perhaps urandom is not supported: %s' % e)
        val = u''.join(u'%02x' % random.randrange(256)
                       for i in range(nbytes))
    return val


def check_output(*args, **kwds):
    """Unicode-aware wrapper for subprocess.check_output"""
    process = subprocess.Popen(stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                               *args, **kwds)
    output, errors = process.communicate()
    retcode = process.poll()
    if errors:
        LOG.debug('ran: %s', ' '.join(*args))
        LOG.debug('returned: %s', retcode)
        LOG.debug('error output: %s', errors.rstrip())
        LOG.debug('regular output: %s', output.rstrip())
    if retcode:
        LOG.debug('raising error')
        raise subprocess.CalledProcessError(retcode, args, output=output)
    return output.decode('utf-8')



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
import textwrap

import mock

from reno import cache
from reno import config
from reno.tests import base


class TestCache(base.TestCase):

    scanner_output = {
        '0.0.0': [('note1', 'shaA')],
        '1.0.0': [('note2', 'shaB'), ('note3', 'shaC')],
    }

    note_bodies = {
        'note1': textwrap.dedent("""
        prelude: >
          This is the prelude.
        """),
        'note2': textwrap.dedent("""
        issues:
          - This is the first issue.
          - This is the second issue.
        """),
        'note3': textwrap.dedent("""
        features:
          - We added a feature!
        """)
    }

    def _get_note_body(self, filename, sha):
        return self.note_bodies.get(filename, '')

    def setUp(self):
        super(TestCache, self).setUp()
        self.useFixture(
            fixtures.MockPatch('reno.scanner.Scanner.get_file_at_commit',
                               new=self._get_note_body)
        )
        self.c = config.Config('.')

    def test_build_cache_db(self):
        with mock.patch('reno.scanner.Scanner.get_notes_by_version') as gnbv:
            gnbv.return_value = self.scanner_output
            db = cache.build_cache_db(
                self.c,
                versions_to_include=[],
            )
            expected = {
                'notes': [
                    {'version': k, 'files': v}
                    for k, v in self.scanner_output.items()
                ],
                'file-contents': {
                    'note1': {
                        'prelude': 'This is the prelude.\n',
                    },
                    'note2': {
                        'issues': [
                            'This is the first issue.',
                            'This is the second issue.',
                        ],
                    },
                    'note3': {
                        'features': ['We added a feature!'],
                    },
                },
            }
            self.assertEqual(expected, db)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
import io
import mock

from reno import create
from reno.tests import base


class TestPickFileName(base.TestCase):

    @mock.patch('os.path.exists')
    def test_not_random_enough(self, exists):
        exists.return_value = True
        self.assertRaises(
            ValueError,
            create._pick_note_file_name,
            'somepath',
            'someslug',
        )

    @mock.patch('os.path.exists')
    def test_random_enough(self, exists):
        exists.return_value = False
        result = create._pick_note_file_name('somepath', 'someslug')
        self.assertIn('somepath', result)
        self.assertIn('someslug', result)


class TestCreate(base.TestCase):

    def setUp(self):
        super(TestCreate, self).setUp()
        self.tmpdir = self.useFixture(fixtures.TempDir()).path

    def _create_user_template(self, contents):
        filename = create._pick_note_file_name(self.tmpdir, 'usertemplate')
        with open(filename, 'w') as f:
            f.write(contents)
        return filename

    def _get_file_path_from_output(self, output):
        # Get the last consecutive word from the output and remove the newline
        return output[output.rfind(" ") + 1:-1]

    def test_create_from_template(self):
        filename = create._pick_note_file_name(self.tmpdir, 'theslug')
        create._make_note_file(filename, 'i-am-a-template')
        with open(filename, 'r') as f:
            body = f.read()
        self.assertEqual('i-am-a-template', body)

    def test_create_from_user_template(self):
        args = mock.Mock()
        args.from_template = self._create_user_template('i-am-a-user-template')
        args.slug = 'theslug'
        args.edit = False
        conf = mock.Mock()
        conf.notespath = self.tmpdir
        with mock.patch('sys.stdout', new=io.StringIO()) as fake_out:
            create.create_cmd(args, conf)
        filename = self._get_file_path_from_output(fake_out.getvalue())
        with open(filename, 'r') as f:
            body = f.read()
        self.assertEqual('i-am-a-user-template', body)

    def test_create_from_user_template_fails_because_unexistent_file(self):
        args = mock.Mock()
        args.from_template = 'some-unexistent-file.yaml'
        args.slug = 'theslug'
        args.edit = False
        conf = mock.Mock()
        conf.notespath = self.tmpdir
        self.assertRaises(ValueError, create.create_cmd, args, conf)

    def test_edit(self):
        self.useFixture(fixtures.EnvironmentVariable('EDITOR', 'myeditor'))
        with mock.patch('subprocess.call') as call_mock:
            self.assertTrue(create._edit_file('somepath'))
            call_mock.assert_called_once_with(['myeditor', 'somepath'])

    def test_edit_without_editor_env_var(self):
        self.useFixture(fixtures.EnvironmentVariable('EDITOR'))
        with mock.patch('subprocess.call') as call_mock:
            self.assertFalse(create._edit_file('somepath'))
            call_mock.assert_not_called()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function


def _indent_for_list(text, prefix='  '):
    """Indent some text to make it work as a list entry.

    Indent all lines except the first with the prefix.
    """
    lines = text.splitlines()
    return '\n'.join([lines[0]] + [
        prefix + l
        for l in lines[1:]
    ]) + '\n'


def format_report(loader, config, versions_to_include, title=None,
                  show_source=True):
    report = []
    if title:
        report.append('=' * len(title))
        report.append(title)
        report.append('=' * len(title))
        report.append('')

    # Read all of the notes files.
    file_contents = {}
    for version in versions_to_include:
        for filename, sha in loader[version]:
            body = loader.parse_note_file(filename, sha)
            file_contents[filename] = body

    for version in versions_to_include:
        if '-' in version:
            # This looks like an "unreleased version".
            title = config.unreleased_version_title or version
        else:
            title = version
        report.append(title)
        report.append('=' * len(title))
        report.append('')

        # Add the preludes.
        notefiles = loader[version]
        prelude_name = config.prelude_section_name
        notefiles_with_prelude = [(n, sha) for n, sha in notefiles
                                  if prelude_name in file_contents[n]]
        if notefiles_with_prelude:
            report.append(prelude_name.replace('_', ' ').title())
            report.append('-' * len(prelude_name))
            report.append('')

        for n, sha in notefiles_with_prelude:
            if show_source:
                report.append('.. %s @ %s\n' % (n, sha))
            report.append(file_contents[n][prelude_name])
            report.append('')

        # Add other sections.
        for section_name, section_title in config.sections:
            notes = [
                (n, fn, sha)
                for fn, sha in notefiles
                if file_contents[fn].get(section_name)
                for n in file_contents[fn].get(section_name, [])
            ]
            if notes:
                report.append(section_title)
                report.append('-' * len(section_title))
                report.append('')
                for n, fn, sha in notes:
                    if show_source:
                        report.append('.. %s @ %s\n' % (fn, sha))
                    report.append('- %s' % _indent_for_list(n))
                report.append('')

    return '\n'.join(report)
